#!C:\Users\builder\mc3-win32\conda-bld\jsonschema_1505949413539\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'jsonschema==2.6.0','console_scripts','jsonschema'
__requires__ = 'jsonschema==2.6.0'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('jsonschema==2.6.0', 'console_scripts', 'jsonschema')()
    )
